/*
 * GNOME Contacts
 */

/* The contacts in the left pane */
.contacts-contact-list {
  background-color: transparent;
}

/* The search entry to filter the list of contacts */
.contacts-filter-entry {
  margin: 6px;
}

/* Draw a little shadow about the contact list when scrolling */
.contact-list-scrolled-window undershoot.top {
  box-shadow: inset 0 1px @borders;
}

/* The link suggestion widget */
.contacts-link-suggestion {
  border-top: 1px solid @borders;
  background-color: shade(@theme_bg_color, 0.9);
}

  .contacts-link-suggestion avatar {
    margin: 12px;
  }

.contacts-avatar-popover .contact-display-name {
  font-size: 20px;
}

flowboxchild.circular {
  padding: 4px;
  border-radius: 9999px;
}

/* Contact Sheet/Editor common */
/* --------------------------- */

.contacts-sheet-container,
.contacts-contact-editor-container {
  margin: 36px 12px 24px 12px;
}

.contacts-sheet-property,
.contacts-editor-property {
  margin: 9px 6px;
}

/* Contact Sheet-related CSS class */
/* ------------------------------- */

.contacts-sheet {
}

.contacts-sheet-header {
  margin: 0 0 12px 0;
}

.contacts-sheet-property {
}


/* Contact Editor-related CSS classes */
/* ---------------------------------- */

/* Common class for all widgets editing a property  */
.contacts-editor-property {
}

  .contacts-editor-property .contacts-property-icon {
    margin: 12px 12px;
  }

  .contacts-editor-property button.contacts-editor-main-widget {
    padding: 10px 0;
  }

.contacts-editor-address,
.contacts-editor-role {
  padding-top: 6px;
  padding-bottom: 6px;
}

/* Widget to edit a birthday (_not_ the row itself) */
.contacts-editor-birthday {
  margin: 12px;
}
